/*********************************************
 * vim:sw=8:ts=8:si:et
 * To use the above modeline in vim you must have "set modeline" in your .vimrc
 * Author: Guido Socher
 * Copyright: GPL V2
 * See http://www.gnu.org/licenses/gpl.html
 *
 * Ethernet weather station with LCD display
 * and HTTP interface 
        url looks like this http://baseurl/
 *
 * Chip type           : ATMEGA88 or ATMEGA168 with ENC28J60
 *********************************************/
#include <avr/io.h>
#include <stdlib.h>
#include <string.h>
#include <avr/pgmspace.h>
#include "ip_arp_udp_tcp.h"
#include "enc28j60.h"
#include "timeout.h"
#include "avr_compat.h"
#include "analog.h"
#include "net.h"
#include "lcd.h"
#include "config.h"

// show debug information about raw ADC values:
//#define SHOWADCRAW 1
//
// please modify the following two lines. mac and ip have to be unique
// in your local area network. You can not have the same numbers in
// two devices:
static uint8_t mymac[6] = {0x54,0x55,0x58,0x10,0x00,0x25};
// how did I get the mac addr? Translate the first 3 numbers into ascii is: TUX
static uint8_t myip[4] = {10,0,0,25};
// base url (you can put a DNS name instead of an IP addr. if you have
// a DNS server (baseurl must end in "/"):
static char baseurl[]="http://10.0.0.25/";
static uint16_t mywwwport =80; // listen port for tcp/www (max range 1-254)
// or on a different port:
//static char baseurl[]="http://10.0.0.25:88/";
//static uint16_t mywwwport =88; // listen port for tcp/www (max range 1-254)
//

#define BUFFER_SIZE 450
static uint8_t buf[BUFFER_SIZE+1];
// local display by default in Celsius or Fahrenheit
static uint8_t celsius=1; // 1=default Celsius

// insert a '.' before the last digit/character in the string.
// The user is responsible for allocating enough space.
// "100" will become "10.0"
void adddecimalpoint(char *str){
        char c;
        while(*str){
                c=*str;
                str++;
                if (!*str){
                        // last char, insert a dot
                        *(str-1)='.';
                        *str=c;
                        str++;
                        *str='\0';
                }
        }
}

void print_lcd()
{
        char numstr[8];
        int i;
        lcd_clrscr();
        lcd_puts_p(PSTR("i:"));
        //--- in-door sensor for temperature
        i=convertanalog(1);// build an average
        i+=convertanalog(1);
        i/=2;
        i=tempcelsius1(i);
        if (celsius){
                itoa(i,numstr,10);
        }else{
                itoa(c2f(i),numstr,10);
        }
        adddecimalpoint(numstr);
        lcd_puts(numstr);
        lcd_putc('\'');
        if (celsius){
                lcd_putc('C');
        }else{
                lcd_putc('F');
        }

        //--- out-door sensor for temperature
        lcd_puts_p(PSTR(" o:"));
        i=convertanalog(0);// build an average
        i+=convertanalog(0);
        i/=2;
        i=tempcelsius0(i);
        if (celsius){
                itoa(i,numstr,10);
        }else{
                itoa(c2f(i),numstr,10);
        }
        adddecimalpoint(numstr);
        lcd_puts(numstr);
        lcd_putc('\'');
        if (celsius){
                lcd_putc('C');
        }else{
                lcd_putc('F');
        }

        lcd_gotoxy(0,1); //next line
        //--- air pressure sensor 
        i=convertanalog(2);// build an average
        i+=convertanalog(2);
        i/=2;
        i=pressure(i); 
        itoa(pressureATsealevel(i),numstr,10);
        adddecimalpoint(numstr);
        lcd_puts(numstr);
        lcd_puts_p(PSTR(" hPa nrm")); // normalized to sea level
}

// prepare the webpage by writing the data to the tcp send buffer
uint16_t print_webpage(uint8_t *buf)
{
        uint16_t plen;
        char numstr[8];
        int i;

        plen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 200 OK\r\nContent-Type: text/html\r\nPragma: no-cache\r\n\r\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<h2>AVR Weather Station</h2>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<pre>\nIn-door temp : "));
        //
        //--- in-door sensor for temperature
        i=convertanalog(1);// build an average
        i+=convertanalog(1);
        i/=2;
#ifdef SHOWADCRAW
        itoa(i,numstr,10); // convert integer to string
        plen=fill_tcp_data(buf,plen,numstr);
        plen=fill_tcp_data_p(buf,plen,PSTR(" "));
#endif
        i=tempcelsius1(i);
        itoa(i,numstr,10);
        adddecimalpoint(numstr);
        plen=fill_tcp_data(buf,plen,numstr);
        plen=fill_tcp_data_p(buf,plen,PSTR("'C ["));
        itoa(c2f(i),numstr,10);
        adddecimalpoint(numstr);
        plen=fill_tcp_data(buf,plen,numstr);
        plen=fill_tcp_data_p(buf,plen,PSTR("'F]"));

        //--- out-door sensor for temperature
        plen=fill_tcp_data_p(buf,plen,PSTR("\nOut-door temp: "));
        i=convertanalog(0);// build an average
        i+=convertanalog(0);
        i/=2;
#ifdef SHOWADCRAW
        itoa(i,numstr,10); // convert integer to string
        plen=fill_tcp_data(buf,plen,numstr);
        plen=fill_tcp_data_p(buf,plen,PSTR(" "));
#endif
        i=tempcelsius0(i);
        itoa(i,numstr,10);
        adddecimalpoint(numstr);
        plen=fill_tcp_data(buf,plen,numstr);
        plen=fill_tcp_data_p(buf,plen,PSTR("'C ["));
        itoa(c2f(i),numstr,10);
        adddecimalpoint(numstr);
        plen=fill_tcp_data(buf,plen,numstr);
        plen=fill_tcp_data_p(buf,plen,PSTR("'F]"));

        //--- air pressure sensor 
        plen=fill_tcp_data_p(buf,plen,PSTR("\nAir pressure : "));
        i=convertanalog(2);// build an average
        i+=convertanalog(2);
        i/=2;
#ifdef SHOWADCRAW
        itoa(i,numstr,10); // convert integer to string
        plen=fill_tcp_data(buf,plen,numstr);
        plen=fill_tcp_data_p(buf,plen,PSTR(" "));
#endif
        i=pressure(i); 
        itoa(i,numstr,10);
        adddecimalpoint(numstr);
        plen=fill_tcp_data(buf,plen,numstr);
        plen=fill_tcp_data_p(buf,plen,PSTR(" hPa real\n               "));
        itoa(pressureATsealevel(i),numstr,10);
        adddecimalpoint(numstr);
        plen=fill_tcp_data(buf,plen,numstr);
        plen=fill_tcp_data_p(buf,plen,PSTR(" hPa at sea level"));
        //
#ifdef SHOWADCRAW
        plen=fill_tcp_data_p(buf,plen,PSTR("\nADC3:"));
        itoa(convertanalog(3),numstr,10); // convert integer to string
        plen=fill_tcp_data(buf,plen,numstr);
#endif
        //
        plen=fill_tcp_data_p(buf,plen,PSTR("</pre>\n<br>"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<small><a href=\""));
        plen=fill_tcp_data(buf,plen,baseurl);
        plen=fill_tcp_data_p(buf,plen,PSTR("\">[refresh]</a></small><br>\n<hr><br>"));
        plen=fill_tcp_data_p(buf,plen,PSTR(VERSIONSTR));
        return(plen);
}


int main(void){

        
        uint16_t plen;
        uint16_t dat_p;
        uint16_t l=0;
        uint8_t toggle=0;
        uint8_t buttonpress=0;
        
        // set the clock speed to "no pre-scaler" (8MHz with internal osc or 
        // full external speed)
        // set the clock prescaler. First write CLKPCE to enable setting of clock the
        // next four instructions.
        CLKPR=(1<<CLKPCE); // change enable
        CLKPR=0; // "no pre-scaler"
        delay_ms(1);

        /* enable PD2/INT0, as input */
        DDRD&= ~(1<<DDD2);

        /*initialize enc28j60*/
        enc28j60Init(mymac);
        enc28j60clkout(2); // change clkout from 6.25MHz to 12.5MHz
        delay_ms(10);
        
        // Push button ( note: in precious circuits this pin was used as LED out)
        // enable PB1 as input
        DDRB&= ~(1<<DDB1);
        // pullup on
        PORTB|= (1<<PINB1);

        /* Magjack leds configuration, see enc28j60 datasheet, page 11 */
        // LEDB=yellow LEDA=green
        //
        // 0x476 is PHLCON LEDA=links status, LEDB=receive/transmit
        // enc28j60PhyWrite(PHLCON,0b0000 0100 0111 01 10);
        enc28j60PhyWrite(PHLCON,0x476);
        delay_ms(20);
        

        // lcd display:
        lcd_init(LCD_DISP_ON);
        lcd_clrscr();
        lcd_puts_P("=ok=");
        delay_ms(20);

        //init the ethernet/ip layer:
        init_ip_arp_udp_tcp(mymac,myip,mywwwport);

        while(1){
                // get the next new packet:
                plen = enc28j60PacketReceive(BUFFER_SIZE, buf);

                /*plen will ne unequal to zero if there is a valid 
                 * packet (without crc error) */
                if(plen==0){
                        l++;
                        if (bit_is_clear(PINB,PINB1)){
                                // push button pressed
                                l=0; // 'l' works also as delay counter for button press
                                     // we need this to de-bounce the button
                                buttonpress=1;

                        }
                        if (l==200){
                                // update local display once in a while
                                if (buttonpress){
                                        buttonpress=0;
                                        celsius=celsius ^ 1; // xor
                                }
                                if (toggle==0){
                                        lcd_clrscr();
                                        print_lcd();
                                        lcd_gotoxy(19,1);
                                        lcd_putc('+');
                                }else{
                                        print_lcd();
                                }
                                toggle=toggle ^ 1; // xor
                        }
                        continue;
                }
                        
                // arp is broadcast if unknown but a host may also
                // verify the mac address by sending it to 
                // a unicast address.
                if(eth_type_is_arp_and_my_ip(buf,plen)){
                        make_arp_answer_from_request(buf);
                        continue;
                }

                // check if ip packets are for us:
                if(eth_type_is_ip_and_my_ip(buf,plen)==0){
                        continue;
                }
                
                if(buf[IP_PROTO_P]==IP_PROTO_ICMP_V && buf[ICMP_TYPE_P]==ICMP_TYPE_ECHOREQUEST_V){
                        // a ping packet, let's send pong
                        make_echo_reply_from_request(buf,plen);
                        continue;
                }
                // tcp port www start, compare only the lower byte
                if (buf[IP_PROTO_P]==IP_PROTO_TCP_V&&buf[TCP_DST_PORT_H_P]==0&&buf[TCP_DST_PORT_L_P]==mywwwport){
                        if (buf[TCP_FLAGS_P] & TCP_FLAGS_SYN_V){
                                make_tcp_synack_from_syn(buf);
                                // make_tcp_synack_from_syn does already send the syn,ack
                                continue;
                        }
                        if (buf[TCP_FLAGS_P] & TCP_FLAGS_ACK_V){
                                init_len_info(buf); // init some data structures
                                // we can possibly have no data, just ack:
                                dat_p=get_tcp_data_pointer();
                                if (dat_p==0){
                                        if (buf[TCP_FLAGS_P] & TCP_FLAGS_FIN_V){
                                                // finack, answer with ack
                                                make_tcp_ack_from_any(buf);
                                        }
                                        // just an ack with no data, wait for next packet
                                        continue;
                                }
                                if (strncmp("GET ",(char *)&(buf[dat_p]),4)!=0){
                                        // head, post and other methods:
                                        //
                                        // for possible status codes see:
                                        // http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html
                                        plen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 200 OK\r\nContent-Type: text/html\r\n\r\n<h1>200 OK</h1>"));
                                        goto SENDTCP;
                                }
                                // any valid url:
                                if (strncmp("/ ",(char *)&(buf[dat_p+4]),2)==0){
                                        plen=print_webpage(buf);
                                }else{
                                // garbage:
                                        plen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 400 Bad Request\r\nContent-Type: text/html\r\n\r\n<h1>400 Bad Request</h1>"));
                                }
SENDTCP:
                                make_tcp_ack_from_any(buf); // send ack for http get
                                make_tcp_ack_with_data(buf,plen); // send data
                                continue;
                        }

                }
                // tcp port www end
                //
        }
        return (0);
}
